# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024 Greenbone AG

from pathlib import Path

from tests.plugins import PluginTestCase
from troubadix.plugin import LinterWarning
from troubadix.plugins.variable_redefinition_in_foreach import (
    CheckVariableRedefinitionInForeach,
)


class CheckVariableRedefinitionInForeachTestCase(PluginTestCase):
    def test_ok(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            'urls = ["foo", "bar"];\n'
            "foreach url(urls) {\n  display(url);\n}\n"
            'url1 = "foo";\n'
            'foreach url(make_list(url1, "bar")) {\n  display(url);\n}'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckVariableRedefinitionInForeach(fake_context)
        results = list(plugin.run())

        self.assertEqual(len(results), 0)

    def test_fail(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            'url1 = "foo";\n'
            "foreach url(url) {\n  display(url);\n}\n"
            "foreach url(make_list(url1, url)) {\n  display(url);\n}\n"
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckVariableRedefinitionInForeach(fake_context)
        results = list(plugin.run())
        self.assertEqual(len(results), 2)
        self.assertIsInstance(results[0], LinterWarning)
        self.assertEqual(
            "The variable 'url' is redefined "
            "by being the identifier\nand the iterator in the"
            " same foreach loop 'foreach url(url)'",
            results[0].message,
        )
        self.assertEqual(
            "The variable 'url' is used as identifier and\n"
            "as part of the iterator in the"
            " same foreach loop\n'foreach url(make_list(url1, url))'",
            results[1].message,
        )



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024 Greenbone AG
# pylint: disable=line-too-long
# pylint: disable=protected-access
import unittest
from pathlib import Path

from tests.plugins import TemporaryDirectory
from troubadix.standalone_plugins.deprecate_vts import (
    DeprecatedFile,
    _finalize_content,
    _get_summary,
    deprecate,
    get_files_from_path,
    parse_args,
    parse_files,
    update_summary,
)


class ParseArgsTestCase(unittest.TestCase):
    def test_parse_args(self):
        testfile = "testfile.nasl"
        output_path = "attic/"
        reason = "NOTUS"

        args = parse_args(
            [
                "--file",
                testfile,
                "--output-path",
                output_path,
                "--deprecation-reason",
                reason,
            ]
        )
        self.assertEqual(args.file, [Path(testfile)])
        self.assertEqual(args.output_path, Path(output_path))
        self.assertEqual(args.deprecation_reason, reason)

    def test_mandatory_arg_group_both(self):
        testfile = "testfile.nasl"
        output_path = "attic/"
        input_path = "nasl/common"
        reason = "NOTUS"

        with self.assertRaises(SystemExit):
            parse_args(
                [
                    "--file",
                    testfile,
                    "--output-path",
                    output_path,
                    "--input-path",
                    input_path,
                    "--deprecation-reason",
                    reason,
                ]
            )

    def test_invalid_reason(self):
        output_path = "attic/"
        input_path = "nasl/common"
        reason = "foo"
        with self.assertRaises(SystemExit):
            parse_args(
                [
                    "--output-path",
                    output_path,
                    "--input-path",
                    input_path,
                    "--deprecation-reason",
                    reason,
                ]
            )

    def test_mandatory_arg_group_neither(self):
        output_path = "attic/"
        reason = "NOTUS"
        with self.assertRaises(SystemExit):
            parse_args(
                [
                    "--output-path",
                    output_path,
                    "--deprecation-reason",
                    reason,
                ]
            )

    def test_parse_from_file(self):
        output_path = "attic/"
        with TemporaryDirectory() as tempdir:
            from_file = tempdir / "from_file.txt"
            testfile1 = tempdir / "testfile1.nasl"
            testfile2 = tempdir / "testfile2.nasl"

            from_file.write_text(f"{testfile1}\n{testfile2}\n", encoding="utf8")

            args = parse_args(
                [
                    "--from-file",
                    str(from_file),
                    "--output-path",
                    output_path,
                    "-r",
                    "NOTUS",
                ]
            )

            self.assertEqual(args.from_file, from_file)


NASL_CONTENT = (
    '...if(description)\n{\n  script_oid("1.3.6.1.4.1.25623.1.0.910673");'
    '\n  script_version("2024-03-12T14:15:13+0000");'
    '\n  script_name("RedHat: Security Advisory for gd (RHSA-2020:5443-01)");'
    '\n  script_family("Red Hat Local Security Checks");\n  script_dependencies("gather-package-list.nasl");'  # noqa: E501
    '\n  script_mandatory_keys("ssh/login/rhel", "ssh/login/rpms", re:"ssh/login/release=RHENT_7");'
    '\n\n  script_xref(name:"RHSA", value:"2020:5443-01");\n  script_xref(name:"URL", value:"https://www.redhat.com/archives/rhsa-announce/2020-December/msg00044.html");'
    '\n\n  script_tag(name:"summary", value:"The remote host is missing an update for the \'gd\'\n  package(s) announced via the RHSA-2020:5443-01 advisory.");'  # noqa: E501
    '\n\n  exit(0);\n}\n\ninclude("revisions-lib.inc");\ninclude("pkg-lib-rpm.inc");\n\nrelease = rpm_get_ssh_release();\nif(!release)\n  exit(0);\n\nres = "";\nreport = "";\n\nif(release == "RHENT_7") {\n\n  if(!isnull(res = isrpmvuln(pkg:"gd", rpm:"gd~2.0.35~27.el7_9", rls:"RHENT_7"))) {\n    report += res;\n  }\n\n  if(!isnull(res = isrpmvuln(pkg:"gd-debuginfo", rpm:"gd-debuginfo~2.0.35~27.el7_9", rls:"RHENT_7"))) {\n    report += res;\n  }\n\n  if(report != "") {\n    security_message(data:report);\n  } else if(__pkg_match) {\n    exit(99);\n  }\n  exit(0);\n}\n\nexit(0);'  # noqa: E501
)

NASL_CONTENT_KB = (
    '...if(description)\n{\n  script_oid("1.3.6.1.4.1.25623.1.0.910673");'
    '\n  script_mandatory_keys("ssh/login/rhel", "ssh/login/rpms", re:"ssh/login/release=RHENT_7");'
    '\n\n  set_kb_item(name:"shttp/" + port + "/detected", value:TRUE);"'
    '\n\n  script_tag(name:"summary", value:"The remote host is missing an update for the \'gd\'\n  package(s) announced via the RHSA-2020:5443-01 advisory.");'  # noqa: E501
    '\n\n  exit(0);\n}\n\ninclude("revisions-lib.inc");\ninclude("pkg-lib-rpm.inc");\n\nrelease = rpm_get_ssh_release();\nif(!release)\n  exit(0);\n\nres = "";\nreport = "";\n\nif(release == "RHENT_7") {\n\n  if(!isnull(res = isrpmvuln(pkg:"gd", rpm:"gd~2.0.35~27.el7_9", rls:"RHENT_7"))) {\n    report += res;\n  }\n\n  if(!isnull(res = isrpmvuln(pkg:"gd-debuginfo", rpm:"gd-debuginfo~2.0.35~27.el7_9", rls:"RHENT_7"))) {\n    report += res;\n  }\n\n  if(report != "") {\n    security_message(data:report);\n  } else if(__pkg_match) {\n    exit(99);\n  }\n  exit(0);\n}\n\nexit(0);'  # noqa: E501
)


class DeprecateVTsTestCase(unittest.TestCase):
    def test_deprecate(self):
        with TemporaryDirectory() as out_dir, TemporaryDirectory() as in_dir:
            testfile1 = in_dir / "testfile1.nasl"
            testfile1.write_text(NASL_CONTENT, encoding="utf8")

            testfile2 = out_dir / "testfile1.nasl"
            testfile2.touch()

            to_deprecate = [
                DeprecatedFile(
                    name="testfile1.nasl",
                    full_path=testfile1,
                    content=NASL_CONTENT,
                )
            ]
            deprecate(out_dir, to_deprecate, "NOTUS")

            result = testfile2.read_text(encoding="utf8")
            self.assertNotIn(result, "script_mandatory_keys")
            self.assertNotIn(result, "script_dependencies")
            self.assertNotIn(result, 'include("revisions-lib.inc");')
            assert (
                "Note: This VT has been deprecated and replaced by "
                "a Notus scanner based one."
            ) in result

    def test_deprecate_kb_item(self):
        with TemporaryDirectory() as out_dir, TemporaryDirectory() as in_dir:
            testfile1 = in_dir / "testfile1.nasl"
            testfile1.write_text(NASL_CONTENT_KB, encoding="utf8")

            to_deprecate = [
                DeprecatedFile(
                    name="testfile1.nasl",
                    full_path=testfile1,
                    content=NASL_CONTENT_KB,
                )
            ]
            deprecate(out_dir, to_deprecate, "NOTUS")
            self.assertLogs(
                "Unable to deprecate testfile1.nasl. There are still KB keys "
                "remaining."
            )

    def test_get_summary(self):
        result = _get_summary(NASL_CONTENT)
        expected = (
            "The remote host is missing an update for the 'gd'\n  package(s) "
            "announced "
            "via the RHSA-2020:5443-01 advisory."
        )
        self.assertEqual(result, expected)

    def test_finalize_content(self):

        result = _finalize_content(NASL_CONTENT)
        expected = (
            '...if(description)\n{\n  script_oid("1.3.6.1.4.1.25623.1.0.910673");\n  '
            'script_version("2024-03-12T14:15:13+0000");\n  script_name("RedHat: Security Advisory for gd (RHSA-2020:5443-01)");\n  script_family("Red Hat Local Security Checks");\n  script_dependencies("gather-package-list.nasl");\n  script_mandatory_keys("ssh/login/rhel", "ssh/login/rpms", re:"ssh/login/release=RHENT_7");\n\n  script_xref(name:"RHSA", value:"2020:5443-01");\n  script_xref(name:"URL", value:"https://www.redhat.com/archives/rhsa-announce/2020-December/msg00044.html");\n\n  script_tag(name:"summary", value:"The remote host is missing an update for the \'gd\'\n  package(s) announced via the RHSA-2020:5443-01 advisory.");\n\n  script_tag(name:"deprecated", value:TRUE);\n\nexit(0);\n}\n\nexit(66);\n'  # noqa: E501
        )
        self.assertEqual(result, expected)

    def test_update_summary_no_oid_match(self):
        file = DeprecatedFile(
            name="testfile.nasl",
            full_path=Path("dir/testfile.nasl"),
            content=NASL_CONTENT,
        )
        result = update_summary(file, "NOTUS")
        self.assertIn("This VT has been deprecated", result)

    def test_get_files_from_path(self):
        with TemporaryDirectory() as in_dir:
            testfile1 = in_dir / "gb_rhsa_2021_8383_8383.nasl"
            testfile1.write_text(NASL_CONTENT, encoding="utf8")

            result = get_files_from_path(dir_path=in_dir)
            expected = [testfile1]

            self.assertEqual(result, expected)

    def test_parse_files(self):
        with TemporaryDirectory() as in_dir:
            testfile1 = in_dir / "gb_rhsa_2021_8383_8383.nasl"
            testfile1.write_text(NASL_CONTENT, encoding="utf8")

            result = parse_files(files=[testfile1])
            expected = [
                DeprecatedFile(
                    name="gb_rhsa_2021_8383_8383.nasl",
                    full_path=in_dir / "gb_rhsa_2021_8383_8383.nasl",
                    content=NASL_CONTENT,
                )
            ]

            self.assertEqual(result, expected)



# SPDX-FileCopyrightText: 2021 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import re
from pathlib import Path
from typing import Iterator

from troubadix.helper import is_ignore_file
from troubadix.helper.patterns import (
    ScriptTag,
    SpecialScriptTag,
    get_script_tag_pattern,
    get_special_script_tag_pattern,
)
from troubadix.plugin import (
    FileContentPlugin,
    LinterError,
    LinterResult,
)

_IGNORE_FILES = ()


SPDX_OR_COPYRIGHT_PATTERN = re.compile(
    r"(SPDX-FileCopyrightText:|Copyright \(C\))\s*(?P<year>\d{4})"
)


def _generate_header_lines(file_content: str):
    for line in file_content.splitlines():
        if line.startswith("#"):
            yield line
        else:
            break


class CheckCopyrightYear(FileContentPlugin):
    """This steps checks if a VT contains a Copyright statement containing a
    year not matching the year defined in the creation_date statement.

    pre2008 script copyright are allowed to predate the creation_year

    For scripts that contain multiple copyrights with different dates,
    like gather-package-list.nasl, it is important that the newer one,
    which is only valid as of a certain year, has information about
    what it covers in the place where the year would normally be.
    Then that line will be automatically ignored
    without having to add it to the ignore list due to not matching the regex.
    Example:
        if creation year = 2008
    ok:
    # SPDX-FileCopyrightText: 2008 Tim Brown
    # SPDX-FileCopyrightText: New detection methods / pattern / code since 2009 Greenbone AG
    not ok / needs ignore entry:
    # New NASL / detection code since 2018 Copyright (C) 2018 Greenbone AG

    """

    name = "check_copyright_year"

    def check_content(
        self, nasl_file: Path, file_content: str
    ) -> Iterator[LinterResult]:
        if nasl_file.suffix == ".inc" or is_ignore_file(
            nasl_file, _IGNORE_FILES
        ):
            return
        # extract creation year from script tag
        creation_date_pattern = get_script_tag_pattern(ScriptTag.CREATION_DATE)
        creation_date_match = creation_date_pattern.search(file_content)
        if not creation_date_match:
            yield LinterError(
                "Missing creation_date statement in VT",
                file=nasl_file,
                plugin=self.name,
            )
            return
        creation_year = int(creation_date_match.group("value")[:4])

        # extract year in value of script_copyright tag
        script_copyright_pattern = get_special_script_tag_pattern(
            SpecialScriptTag.COPYRIGHT
        )
        script_copyright_match = script_copyright_pattern.search(file_content)
        if not script_copyright_match:
            yield LinterError(
                "Missing copyright tag in VT",
                file=nasl_file,
                plugin=self.name,
            )
            return
        copyright_tag_value = script_copyright_match.group("value")
        copyright_tag_match = SPDX_OR_COPYRIGHT_PATTERN.search(
            copyright_tag_value
        )
        if not copyright_tag_match:
            yield LinterError(
                "Unable to extract year from script_copyright tag in VT",
                file=nasl_file,
                plugin=self.name,
            )
            return
        copyright_tag_year = int(copyright_tag_match.group("year"))

        # list comprehension to collect copy right matches from header
        header_lines = _generate_header_lines(file_content)
        header_copyright_years = [
            int(match.group("year"))
            for line in header_lines
            if (match := SPDX_OR_COPYRIGHT_PATTERN.search(line))
        ]

        if not header_copyright_years:
            yield LinterError(
                "VT header is missing a copyright text",
                file=nasl_file,
                plugin=self.name,
            )

        # check if the copyrights years match the creation year
        # pre2008 VTs are allowed to have a copyright that predates the year of creation.
        if copyright_tag_year != creation_year:
            if "pre2008" in str(nasl_file):
                if copyright_tag_year > creation_year:
                    yield LinterError(
                        "a pre2008 vt has a copyright tag year newer than the creation_year",
                        file=nasl_file,
                        plugin=self.name,
                    )
            else:
                yield LinterError(
                    "script_copyright tag does not match the creation year",
                    file=nasl_file,
                    plugin=self.name,
                )
        for copyright_year in header_copyright_years:
            if creation_year != copyright_year:
                if "pre2008" in str(nasl_file):
                    if copyright_year > creation_year:
                        yield LinterError(
                            "a pre2008 vt has a copyright value in the fileheader"
                            " newer than the creation_year",
                            file=nasl_file,
                            plugin=self.name,
                        )
                else:
                    yield LinterError(
                        "a copyright in the fileheader does not match the creation year",
                        file=nasl_file,
                        plugin=self.name,
                    )



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024 Greenbone AG

from pathlib import Path

from tests.plugins import PluginTestCase
from troubadix.helper import CURRENT_ENCODING
from troubadix.plugin import LinterFix
from troubadix.plugins.spaces_before_dots import CheckSpacesBeforeDots


class TestSpacesBeforeDots(PluginTestCase):

    def test_ok(self):
        nasl_file = Path("/some/fake/directory/test.nasl")
        content = """
              script_tag(name:"summary", value:"Foo Bar.");
              script_tag(name:"solution", value:"Foo .NET.");
              script_tag(name:"insight", value:"Foo Bar ...");
            """
        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckSpacesBeforeDots(fake_context)
        results = list(plugin.run())
        self.assertEqual(len(results), 0)

    def test_fail(self):
        nasl_file = Path("/some/fake/directory/test.nasl")
        content = (
            """
              script_tag(name:"summary", value:"Foo Bar .");
              script_tag(name:"vuldetect", value:"Foo Bar .");
              script_tag(name:"insight", value:"Foo Bar .");
              script_tag(name:"impact", value:"Foo . Bar . Foo");
            """
            'script_tag(name:"affected", value:"Foo\n.\nBar.");'
            'script_tag(name:"solution", value:"Foo Bar\n.");'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckSpacesBeforeDots(fake_context)
        results = list(plugin.run())
        self.assertEqual(len(results), 7)
        self.assertEqual(
            results[0].message,
            "value of script_tag summary has at least one occurence of excess"
            " whitespace before a dot:\n"
            " 'script_tag(name:"
            '"summary", value:"Foo Bar .");'
            "'",
        )

    def test_ignore(self):
        nasl_file = Path("/some/fake/directory/test.inc")
        fake_context = self.create_file_plugin_context(nasl_file=nasl_file)
        plugin = CheckSpacesBeforeDots(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 0)

    def test_fix(self):
        with self.create_directory() as tempdir:
            path = tempdir / "file.nasl"
            content = (
                """
                  script_tag(name:"summary", value:"Foo Bar .");
                  script_tag(name:"vuldetect", value:"Foo Bar .");
                  script_tag(name:"insight", value:"Foo .Net Bar .");
                  script_tag(name:"impact", value:"Foo . Bar . Foo");
                """
                'script_tag(name:"affected", value:"Foo\n.\nBar.");'
                'script_tag(name:"solution", value:"Foo Bar\n.");'
            )
            expected_modified_content = (
                """
                  script_tag(name:"summary", value:"Foo Bar.");
                  script_tag(name:"vuldetect", value:"Foo Bar.");
                  script_tag(name:"insight", value:"Foo .Net Bar.");
                  script_tag(name:"impact", value:"Foo. Bar. Foo");
                """
                'script_tag(name:"affected", value:"Foo.\nBar.");'
                'script_tag(name:"solution", value:"Foo Bar.");'
            )
            path.write_text(content, encoding=CURRENT_ENCODING)

            fake_context = self.create_file_plugin_context(
                nasl_file=path, file_content=content
            )

            plugin = CheckSpacesBeforeDots(fake_context)

            # keep list() to consume the iterator
            list(plugin.run())

            results = list(plugin.fix())
            self.assertEqual(len(results), 1)

            self.assertIsInstance(results[0], LinterFix)
            modified_content = path.read_text(encoding=CURRENT_ENCODING)
            self.assertEqual(modified_content, expected_modified_content)



#  Copyright (c) 2022 Greenbone AG
#
#  SPDX-License-Identifier: GPL-3.0-or-later
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Iterator

from troubadix.helper.patterns import ScriptTag, get_script_tag_pattern
from troubadix.plugin import FilePlugin, LinterError, LinterResult

VALID_QOD_NUM_VALUES = [
    "1",
    "30",
    "50",
    "70",
    "75",
    "80",
    "95",
    "97",
    "98",
    "99",
    "100",
]

VALID_QOD_TYPES = [
    "exploit",
    "remote_vul",
    "remote_app",
    "package",
    "registry",
    "remote_active",
    "remote_banner",
    "executable_version",
    "remote_analysis",
    "remote_probe",
    "remote_banner_unreliable",
    "executable_version_unreliable",
    "general_note",
    "package_unreliable",
]

# needed due to script_tag_pattern value not including the quotes
QOD_WITH_QUOTES_PATTERN = re.compile(
    r'script_tag\(\s*name\s*:\s*(?P<quote>[\'"])qod(?P=quote)\s*,'
    r'\s*value\s*:\s*(?P<value_with_quotes>[\'"]?(?P<value>.+?)[\'"]?)\s*\)\s*;'
)


class CheckQod(FilePlugin):
    name = "check_qod"

    def run(self) -> Iterator[LinterResult]:
        """
        The script checks the passed VT for the existence / validity of its QoD
        """

        if self.context.nasl_file.suffix == ".inc":
            return

        file_content = self.context.file_content

        if "# troubadix: disable=template_nd_test_files_fps" in file_content:
            return

        qod_type_pattern = get_script_tag_pattern(ScriptTag.QOD_TYPE)

        match_qod = list(QOD_WITH_QUOTES_PATTERN.finditer(file_content))
        match_qod_type = list(qod_type_pattern.finditer(file_content))

        num_matches = len(match_qod) + len(match_qod_type)
        if num_matches < 1:
            yield LinterError(
                "VT is missing QoD or QoD type",
                file=self.context.nasl_file,
                plugin=self.name,
            )
        if num_matches > 1:
            yield LinterError(
                "VT contains multiple QoD values",
                file=self.context.nasl_file,
                plugin=self.name,
            )

        for match in match_qod:
            full_match = match.group(0)
            full_value = match.group("value_with_quotes")
            value = match.group("value")

            # Check if the value is enclosed in double quotes
            if full_value.startswith('"') and full_value.endswith('"'):

                # Compare against valid values
                if value not in VALID_QOD_NUM_VALUES:
                    yield LinterError(
                        f"Invalid QOD value '{value}' in {full_match}."
                        " Allowed are"
                        f" {', '.join(x for x in VALID_QOD_NUM_VALUES)}",
                        file=self.context.nasl_file,
                        plugin=self.name,
                    )
            else:
                yield LinterError(
                    f"QOD value not properly enclosed in double quotes in {full_match}",
                    file=self.context.nasl_file,
                    plugin=self.name,
                )

        for match in match_qod_type:
            val = match.group("value").replace('"', "")
            if val not in VALID_QOD_TYPES:
                yield LinterError(
                    f"{match.group(0)}: '{match.group('value')}' is an invalid"
                    f" QoD type. Allowed are {', '.join(VALID_QOD_TYPES)}",
                    file=self.context.nasl_file,
                    plugin=self.name,
                )



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024 Greenbone AG

import re
from argparse import ArgumentParser, ArgumentTypeError, Namespace
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Iterable, Optional

from pontos.terminal.terminal import ConsoleTerminal

from troubadix.argparser import directory_type, file_type
from troubadix.helper.patterns import (
    ScriptTag,
    SpecialScriptTag,
    get_script_tag_pattern,
    get_special_script_tag_pattern,
)
from troubadix.troubadix import from_file


class Deprecations(Enum):
    NOTUS = "and replaced by a Notus scanner based one."
    MERGED = "because it has been merged into a different VT."
    DEFUNCT = "and is therefore no longer functional."
    DUPLICATE = "as a duplicate."


@dataclass
class DeprecatedFile:
    name: str
    full_path: Path
    content: str


KB_ITEMS_PATTERN = re.compile(r"set_kb_item\(.+\);")


def existing_file_type(string: str) -> Path:
    file_path = Path(string)
    if not file_path.exists():
        raise ArgumentTypeError(f'File "{string}" does not exist.')
    if not file_path.is_file():
        raise ArgumentTypeError(f'"{string}" is not a file.')
    return file_path


def update_summary(file: DeprecatedFile, deprecation_reason: str) -> str:
    """Update the summary of the nasl script by adding the information
    that the script has been deprecated, and if possible, the oid of
    the new notus script replacing it.

    Args:
        file: DeprecatedFile object containing the content of the VT
        deprecation_reason: The reason this VT is being deprecated,
            from a list of options.

    Returns:
        The updated content of the file
    """
    old_summary = _get_summary(file.content)
    if not old_summary:
        print(f"No summary in: {file.name}")
        return file.content

    deprecate_text = (
        f"Note: This VT has been deprecated "
        f"{Deprecations[deprecation_reason].value}"
    )

    new_summary = old_summary + "\n" + deprecate_text
    file.content = file.content.replace(old_summary, new_summary)

    return file.content


def _finalize_content(content: str) -> str:
    """Update the content field of the nasl script by adding the
    deprecated tag and removing the extra content."""
    content_to_keep = content.split("exit(0);")[0]
    return content_to_keep + (
        'script_tag(name:"deprecated", value:TRUE);'
        "\n\nexit(0);\n}\n\nexit(66);\n"
    )


def get_files_from_path(dir_path: Path = None) -> list:
    """Get a list of files from the input path provided

    Args:
        dir_path (optional): The path to the directory with the files to
            be deprecated
    """
    return [file for file in dir_path.glob("**/*")]


def parse_files(files: list) -> list[DeprecatedFile]:
    """Convert filepaths into
    DeprecatedFile objects

    Args:
        files: a list of files to deprecate, should be .nasl VTs

    Returns:
        List of DeprecatedFile objects

    """
    to_deprecate = []

    for file in files:
        with file.open("r", encoding="latin-1") as fh:
            to_deprecate.append(
                DeprecatedFile(
                    file.name,
                    file.absolute(),
                    fh.read(),
                )
            )
    return to_deprecate


def _get_summary(content: str) -> Optional[str]:
    """Extract the summary from the nasl script"""
    pattern = get_script_tag_pattern(ScriptTag.SUMMARY)
    if match_summary := re.search(pattern, content):
        value = match_summary.group().split('value:"')[1]
        return value.replace('");', "")
    return None


def deprecate(
    output_path: Path,
    to_deprecate: list[DeprecatedFile],
    deprecation_reason: str,
) -> None:
    """Deprecate the selected VTs by removing unnecessary keys, updating the
    summary, and adding the deprecated tag.

    Args:
        output_path: the directory where the deprecated VTs should be written
            to, i.e. "attic"
        to_deprecate: the list of files to be deprecated
        deprecation_reason: The reason this VT is being deprecated,
        from a list of options.
    """
    output_path.mkdir(parents=True, exist_ok=True)
    for file in to_deprecate:
        if re.findall(KB_ITEMS_PATTERN, file.content):
            print(
                f"Unable to deprecate {file.name}. There are still KB keys "
                f"remaining."
            )
            continue
        file.content = update_summary(file, deprecation_reason)
        file.content = _finalize_content(file.content)

        # Drop any unnecessary script tags like script_dependencies(),
        # script_require_udp_ports() or script_mandatory_keys()
        tags_to_remove = list()

        if dependencies := re.search(
            get_special_script_tag_pattern(SpecialScriptTag.DEPENDENCIES),
            file.content,
        ):
            tags_to_remove.append(dependencies.group())

        if udp := re.search(
            get_special_script_tag_pattern(SpecialScriptTag.REQUIRE_UDP_PORTS),
            file.content,
        ):
            tags_to_remove.append(udp.group())

        if man_keys := re.search(
            get_special_script_tag_pattern(SpecialScriptTag.MANDATORY_KEYS),
            file.content,
        ):
            tags_to_remove.append(man_keys.group())

        for tag in tags_to_remove:
            file.content = file.content.replace("  " + tag + "\n", "")

        file.full_path.rename(output_path / file.name)

        with open(output_path / file.name, "w", encoding="latin-1") as f:
            f.write(file.content)
            f.truncate()


def parse_args(args: Iterable[str] = None) -> Namespace:
    parser = ArgumentParser(description="Deprecate VTs")
    parser.add_argument(
        "-o",
        "--output-path",
        metavar="<output_path>",
        type=directory_type,
        required=True,
        help="Path where the deprecated files should be written to.",
    )
    parser.add_argument(
        "-r",
        "--deprecation-reason",
        metavar="<deprecation_reason>",
        choices=[reason.name for reason in Deprecations],
        type=str,
        help="The reason the VT is being deprecated. Options are 'notus':"
        "The VT has been replaced by a new Notus VT. 'Merged': the VT has"
        "been merged with another still active VT, 'duplicate': The VT has"
        "a still active duplicate, 'defunct': The VT is no longer "
        "functional.",
        required=True,
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-f",
        "--file",
        metavar="<file>",
        nargs="+",
        default=None,
        type=file_type,
        help="File to deprecate",
    )
    group.add_argument(
        "-i",
        "--input-path",
        metavar="<input_path>",
        default=None,
        type=directory_type,
        help="Path to a directory where all files should be deprecated.",
    )
    group.add_argument(
        "--from-file",
        metavar="<from_file>",
        default=None,
        type=existing_file_type,
        help=(
            "Path to a single file that contains a list of files "
            "to be deprecated, separated by new lines."
        ),
    )
    return parser.parse_args(args)


def main():
    args = parse_args()
    terminal = ConsoleTerminal()
    input_path = args.input_path if args.input_path else None
    files = []

    if input_path:
        files = get_files_from_path(input_path)
    elif args.from_file:
        files = from_file(include_file=args.from_file, term=terminal)
    elif args.file:
        files = args.file

    to_be_deprecated = parse_files(files)
    deprecate(args.output_path, to_be_deprecated, args.deprecation_reason)


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from pathlib import Path
from typing import Iterable, Iterator

import charset_normalizer

from troubadix.plugin import LineContentPlugin, LinterError, LinterResult

# Only the ASCII and extended ASCII for now... # https://www.ascii-code.com/
# CHAR_SET = r"[^\x00-\xFF]"
# Temporary only check for chars in between 7f-9f, like in the old Feed-QA...
CHAR_SET = r"[\x7F-\x9F]"

ALLOWED_ENCODINGS = ["ascii", "latin_1"]


class CheckEncoding(LineContentPlugin):
    name = "check_encoding"

    def check_lines(
        self,
        nasl_file: Path,
        lines: Iterable[str],
    ) -> Iterator[LinterResult]:
        match = charset_normalizer.from_path(
            nasl_file, threshold=0.4, cp_isolation=ALLOWED_ENCODINGS
        ).best()

        if not match:
            yield LinterError(
                f"VT uses a wrong encoding. "
                f"Allowed encodings are {', '.join(ALLOWED_ENCODINGS)}.",
                file=nasl_file,
                plugin=self.name,
            )

        for index, line in enumerate(lines, 1):
            encoding = re.search(CHAR_SET, line)
            if encoding is not None:
                yield LinterError(
                    "Found invalid character",
                    file=nasl_file,
                    plugin=self.name,
                    line=index,
                )



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024 Greenbone AG
import re
from collections.abc import Iterator
from pathlib import Path

from troubadix.helper import CURRENT_ENCODING
from troubadix.helper.helper import is_ignore_file
from troubadix.helper.patterns import (
    ScriptTag,
    get_script_tag_pattern,
)
from troubadix.plugin import (
    FileContentPlugin,
    LinterFix,
    LinterResult,
    LinterWarning,
)

TAGS = [
    ScriptTag.SUMMARY,
    ScriptTag.VULDETECT,
    ScriptTag.INSIGHT,
    ScriptTag.IMPACT,
    ScriptTag.AFFECTED,
    ScriptTag.SOLUTION,
]

# Regex pattern to match:
# 1. A dot preceded and/or followed by any whitespace character (floating between words)
# 2. A dot preceded by any whitespace character at the end of the string
PATTERN = re.compile(r"\s+\.(\s|$)")
IGNORE = [
    # 21.04 and 22.04 are generated and should not be touched manually
    "21.04/",
    "22.04/",
    # uses dots for beginning of entry in enumeration
    "common/2008/debian/deb_246.nasl",
    "common/2008/debian/deb_266.nasl",
    "common/2008/freebsd/freebsd_5e92e8a2.nasl",
    "common/2008/freebsd/freebsdsa_cpio.nasl",
    "common/2008/freebsd/freebsdsa_cvs2.nasl",
    "common/2009/osc_photoGallery_sql_injection.nasl",
    "common/2009/secpod_novell_edir_mult_vuln_jul09_lin.nasl",
    "common/2009/secpod_novell_edir_mult_vuln_jul09_win.nasl",
    "common/2010/freebsd/freebsd_3a7c5fc4.nasl",
    "common/2012/freebsd/freebsd_a4a809d8.nasl",
    "common/2015/amazon/alas-2014-455.nasl",
    "common/2015/gb_mozilla_firefox_mult_vuln01_mar15_macosx.nasl",
    "common/2015/gb_mozilla_firefox_mult_vuln01_mar15_win.nasl",
    "common/2015/oracle/ELSA-2009-1619.nasl",
    "common/2015/oracle/ELSA-2011-0586.nasl",
    "common/2016/gb_perl_privilege_escalation_vuln_win.nasl",
    "common/2021/dropbear/gb_dropbear_ssh_filename_vuln_may20.nasl",
    "common/2021/eclipse/gb_jetty_GHSA-v7ff-8wcx-gmc5_lin.nasl",
    "common/2021/eclipse/gb_jetty_GHSA-v7ff-8wcx-gmc5_win.nasl",
    "common/gsf/2009/mandriva/gb_mandriva_MDVSA_2008_140.nasl",
    "common/gsf/2009/mandriva/gb_mandriva_MDVSA_2008_141.nasl",
    "common/gsf/2010/mandriva/gb_mandriva_MDVA_2010_173.nasl",
    "common/gsf/2010/mandriva/gb_mandriva_MDVSA_2010_155.nasl",
    "common/gsf/2010/mandriva/gb_mandriva_MDVSA_2010_155_1.nasl",
    "common/gsf/2010/mandriva/gb_mandriva_MDVSA_2010_167.nasl",
    "common/gsf/2020/f5/gb_f5_big_ip_K11315080.nasl",
    "common/gsf/2020/f5/gb_f5_big_iq_K11315080.nasl",
    "common/2022/opensuse/gb_opensuse_2022_1548_1.nasl",
    "common/2024/opensuse/gb_opensuse_2023_3247_1.nasl",
    "common/attic/debian/deb_232_1.nasl",
]


class CheckSpacesBeforeDots(FileContentPlugin):
    name = "check_spaces_before_dots"

    def check_content(
        self, nasl_file: Path, file_content: str
    ) -> Iterator[LinterResult]:
        """
        This plugin checks for excess whitespace before a dot
        in script_tags that have full sentence values
        """
        self.matches = []
        if nasl_file.suffix == ".inc" or is_ignore_file(nasl_file, IGNORE):
            return
        for tag in TAGS:
            pattern = get_script_tag_pattern(tag)
            match = pattern.search(file_content)
            if not match:
                continue

            value = match.group("value")
            value_start = match.start("value")

            for excess_match in PATTERN.finditer(value):
                whitespace_pos = excess_match.start() + value_start
                self.matches.append((whitespace_pos, excess_match.group()))
                fullmatch = match.group()
                yield LinterWarning(
                    f"value of script_tag {match.group('name')} has at least"
                    " one occurence of excess whitespace before a dot:"
                    f"\n '{fullmatch}'",
                    file=nasl_file,
                    plugin=self.name,
                )

    def fix(self) -> Iterator[LinterResult]:

        if not self.matches:
            return

        # Sort matches by position, descending order to avoid messing up indices during replacement
        self.matches.sort(reverse=True)

        file_content = self.context.file_content
        for pos, match_str in self.matches:
            # Replace the match by removing the excess whitespace before the dot
            fixed_str = re.sub(r"\s+\.", ".", match_str)
            file_content = (
                file_content[:pos]
                + fixed_str
                + file_content[pos + len(match_str) :]
            )

        with open(self.context.nasl_file, "w", encoding=CURRENT_ENCODING) as f:
            f.write(file_content)

        yield LinterFix(
            "Excess spaces were removed",
            file=self.context.nasl_file,
            plugin=self.name,
        )



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024 Greenbone AG

import re
from pathlib import Path
from typing import Iterator

from troubadix.plugin import FileContentPlugin, LinterResult, LinterWarning

FOREACH_PATTERN = re.compile(r"foreach\s+(?P<ident>\w+)\s*\((?P<iter>.+)\)")
MAKE_LIST_PATTERN = re.compile(
    r"^(?:make_list|make_list_unique)\((?P<params>.+)\)$"
)


class CheckVariableRedefinitionInForeach(FileContentPlugin):
    name = "check_variable_redefinition_in_foreach"

    def check_content(
        self, nasl_file: Path, file_content: str
    ) -> Iterator[LinterResult]:
        """This plugin checks for a redefinition of the variable
        that is passed to the foreach loop.
        This can be caused by using same variable name
        for both the list and the element being iterated over.
        Incorrect uses of foreach loops that are covered by this plugin are:
        foreach foo(foo){}
        foreach foo(make_list(bar,foo)){}
        foreach foo(make_list_unique(bar,foo)){}
        """

        for foreach_match in FOREACH_PATTERN.finditer(file_content):
            identifier = foreach_match.group("ident")

            # replace instead of strip, because the iterator can contain
            # whitespace as part of the expression.
            # Saving from further strip calls
            # make_list( bar , foo )
            iterator = foreach_match.group("iter").replace(" ", "")

            if make_list_match := MAKE_LIST_PATTERN.fullmatch(iterator):
                make_list_params = make_list_match.group("params").split(",")
                if identifier in make_list_params:
                    yield LinterWarning(
                        f"The variable '{identifier}' "
                        f"is used as identifier and\n"
                        f"as part of the iterator in the"
                        f" same foreach loop\n'{foreach_match.group()}'",
                        plugin=self.name,
                        file=nasl_file,
                    )
            else:
                if identifier == iterator:
                    yield LinterWarning(
                        f"The variable '{identifier}' is redefined "
                        f"by being the identifier\nand the iterator in the"
                        f" same foreach loop '{foreach_match.group()}'",
                        plugin=self.name,
                        file=nasl_file,
                    )



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path

from troubadix.plugins.copyright_year import CheckCopyrightYear

from . import PluginTestCase


class CheckCopyrightYearTestCase(PluginTestCase):
    def test_ok_new_header(self):
        path = Path("some/file.nasl")
        content = (
            "# SPDX-FileCopyrightText: 2022 Greenbone AG\n"
            "# SPDX-FileCopyrightText: 2022 some other person\n"
            '  script_tag(name:"creation_date", value:"2022-05-14 11:24:55 '
            '+0200 (Tue, 14 May 2022)");\n'
            '  script_copyright("Copyright (C) 2022 Greenbone AG");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 0)

    def test_ok_old_header(self):
        path = Path("some/file.nasl")
        content = (
            "# Copyright (C) 2022 Greenbone AG\n"
            "# Copyright (C) 2022 some other Person\n"
            '  script_tag(name:"creation_date", value:"2022-05-14 11:24:55 '
            '+0200 (Tue, 14 May 2022)");\n'
            '  script_copyright("Copyright (C) 2022 Greenbone AG");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 0)

    def test_pre2008_ok_new_header(self):
        # tests that special cases pass:
        # copyright < creation year for pre2008
        # additional copyright that is newer but ok
        path = Path("some/pre2008/file.nasl")
        content = (
            "# SPDX-FileCopyrightText: 2010 Greenbone AG\n"
            "# SPDX-FileCopyrightText: New code since 2022 Greenbone AG\n"
            '  script_tag(name:"creation_date", value:"2022-05-14 11:24:55 '
            '+0200 (Tue, 14 May 2022)");\n'
            '  script_copyright("Copyright (C) 2020 Greenbone AG");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 0)

    def test_pre2008_ok_old_header(self):
        path = Path("some/pre2008/file.nasl")
        content = (
            "# Copyright (C) 2020 Greenbone AG\n"
            '  script_tag(name:"creation_date", value:"2022-05-14 11:24:55 '
            '+0200 (Tue, 14 May 2022)");\n'
            '  script_copyright("Copyright (C) 2020 Greenbone AG");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 0)

    def test_exclude_inc_file(self):
        path = Path("some/file.inc")
        fake_context = self.create_file_plugin_context(nasl_file=path)
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 0)

    def test_missing_creation_date(self):
        path = Path("some/file.nasl")
        content = (
            "# SPDX-FileCopyrightText: 2022 Greenbone AG\n"
            '  script_copyright("Copyright (C) 2022 Greenbone AG");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 1)
        self.assertEqual(
            "Missing creation_date statement in VT", results[0].message
        )

    def test_missing_copyright_tag(self):
        path = Path("some/file.nasl")
        content = (
            "# SPDX-FileCopyrightText: 2022 Greenbone AG\n"
            '  script_tag(name:"creation_date", value:"2022-05-14 11:24:55 '
            '+0200 (Tue, 14 May 2022)");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 1)
        self.assertEqual("Missing copyright tag in VT", results[0].message)

    def test_regex_fail_copyright_tag(self):
        path = Path("some/file.nasl")
        content = (
            "# SPDX-FileCopyrightText: 2022 Greenbone AG\n"
            '  script_tag(name:"creation_date", value:"2022-05-14 11:24:55 '
            '+0200 (Tue, 14 May 2022)");\n'
            '  script_copyright("Copyright (C) Greenbone AG");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 1)
        self.assertEqual(
            "Unable to extract year from script_copyright tag in VT",
            results[0].message,
        )

    def test_missing_header_copyright(self):
        path = Path("some/file.nasl")
        content = (
            '  script_tag(name:"creation_date", value:"2022-05-14 11:24:55 '
            '+0200 (Tue, 14 May 2022)");\n'
            '  script_copyright("Copyright (C) 2022 Greenbone AG");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 1)
        self.assertEqual(
            "VT header is missing a copyright text",
            results[0].message,
        )
        return

    def test_pre2008_fail(self):
        path = Path("some/pre2008/file.nasl")
        content = (
            "# Copyright (C) 2021 Greenbone AG\n"
            '  script_tag(name:"creation_date", value:"2020-05-14 11:24:55 '
            '+0200 (Tue, 14 May 2020)");\n'
            '  script_copyright("Copyright (C) 2021 Greenbone AG");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 2)
        self.assertEqual(
            "a pre2008 vt has a copyright tag year newer than the creation_year",
            results[0].message,
        )
        self.assertEqual(
            "a pre2008 vt has a copyright value in the fileheader"
            " newer than the creation_year",
            results[1].message,
        )

    def test_fail(self):
        path = Path("some/file.nasl")
        content = (
            "# SPDX-FileCopyrightText: 3000 Greenbone AG\n"
            '  script_tag(name:"creation_date", value:"2022-05-14 11:24:55 '
            '+0200 (Tue, 14 May 2022)");\n'
            '  script_copyright("Copyright (C) 1000 Greenbone AG");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckCopyrightYear(fake_context)

        results = list(plugin.run())
        self.assertEqual(len(results), 2)
        self.assertEqual(
            "script_copyright tag does not match the creation year",
            results[0].message,
        )
        self.assertEqual(
            "a copyright in the fileheader does not match the creation year",
            results[1].message,
        )
